import subprocess

def subprocess_execution(args, logger, outfile = None):
    """
    Exécute un sous-processus (commande système)

    Parameters
    ----------
    args: [str]
        arguments pour l'exécution d'un sous-processus, avec en premier argument
        le nom de la commande

    logger: logging.Logger
    """
    try:
        str_args = [str(arg) for arg in args]
        logger.info('Subprocess: \"' + " ".join(str_args) + '\"')
        if outfile is not None:
            with open(outfile, "w") as out:
                process = subprocess.Popen(
                str_args,
                stdout=out,
                stderr=subprocess.STDOUT,
            )

        else:
            process = subprocess.Popen(
                str_args,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
            )
            process_output, _ =  process.communicate()
            logger.info(process_output.decode("utf-8"))


    except (OSError, subprocess.CalledProcessError) as exception:
        logger.info('Exception occured: ' + str(exception))
        logger.info('Subprocess failed')
        return False
    else:
        # no exception was raised
        logger.info('Subprocess finished')



from r2gg._output_costs_from_costs_config import compute_operations_string

def build_lua(costs_config, output_cost_name):
    """
    Fonction qui crée un fichier .lua de profil OSRM à partir d'un dictionnaire de
    définition des coûts (costs_config) et d'un nom de coût spécifique
    (output_cost_name) qui doît être présent dans la valeur associée à la clé
    outputs de costs_config.

    Parameters
    ----------
    costs_config: dict
        Dictionnaire de configuration des coûts correspondant au schéma JSON de
        configuration des coûts
    output_cost_name: str
        Nom de coût spécifique qui doît être présent dans la valeur associée à
        la clé outputs de costs_config.

    Returns
    -------
    str
        chaîne de caractères correspondant à un fichier .lua de profil OSRM pour
        le coût spécifié
    """

    result = "api_version = 4\n"

    output_cost = None
    for output in costs_config["outputs"]:
        if output["name"] == output_cost_name:
            output_cost = output

    assert output_cost != None, "output cost name not in configuration outputs"

    if output_cost["cost_type"]:
        result += _build_setup(bool(output_cost["turn_restrictions"]),
                               output_cost["cost_type"])
    else:
        result += _build_setup(bool(output_cost["turn_restrictions"]))

    result += _build_process_node()
    result += _build_process_way(costs_config, output_cost)
    result += _build_return()
    return result


def _build_setup(turn_res, cost_type = 'duration'):
    """
    Fonction de construction de la fonction setup du fichier .lua

    Parameters
    ----------
    turn_res: bool
        true si on utilise les non_communications, false sinon
    cost_type: str
        type de coût comme défini dans la documentation des profils .lua OSRM
        pour l'instant, parmi {'duration', 'distance'}

    Returns
    -------
    str
        chaîne de caractères correspondant à la fonction setup du fichier .lua
    """

    setup_string = "function setup()\n"
    setup_string += "    return {\n"
    setup_string += "        properties = {\n"
    setup_string += "            use_turn_restrictions = " + str(turn_res).lower() + ",\n"
    # Demi-tour aux wapoints si pas de turn restriction (cf #37043)
    if (not turn_res):
        setup_string += "            continue_straight_at_waypoint = false,\n"
    setup_string += "            weight_name = \'" + cost_type + "\',\n"
    setup_string += "        },\n"
    setup_string += "        classes = { \"toll\", \"bridge\", \"tunnel\" },\n"
    setup_string += "        excludable = {\n"
    setup_string += "            { [\"toll\"] = true },\n"
    setup_string += "            { [\"bridge\"] = true },\n"
    setup_string += "            { [\"tunnel\"] = true },\n"
    setup_string += "            { [\"toll\"] = true, [\"bridge\"] = true },\n"
    setup_string += "            { [\"toll\"] = true, [\"tunnel\"] = true },\n"
    setup_string += "            { [\"bridge\"] = true, [\"tunnel\"] = true },\n"
    setup_string += "            { [\"toll\"] = true, [\"bridge\"] = true, [\"tunnel\"] = true }\n"
    setup_string += "        }\n"
    setup_string += "    }\n"
    setup_string += "end\n"

    return setup_string

def _build_process_node():
    """
    Fonction de construction de la fonction process_node du fichier .lua
    pour l'instant, elle est vide

    Returns
    -------
    str
        chaîne de caractères correspondant à la fonction process_node du fichier .lua
        pour l'instant, elle est vide
    """

    process_node_string = (
        "function process_node (profile, node, result)\n"
        "\n"
        "end\n"
    )

    return process_node_string

def _build_process_way(costs_config, output_cost):
    """
    Fonction de construction de la fonction process_way du fichier .lua

    Parameters
    ----------
    costs_config: dict
        Dictionnaire de configuration des coûts correspondant au schéma JSON de
        configuration des coûts
    output_cost: dict
        Dictionnaire de la configuration du coût dont le nom est output_cost_name

    Returns
    -------
    str
        chaîne de caractères correspondant à la fonction process_way du fichier .lua
    """

    process_way_string = (
        "function process_way (profile, way, result)\n"
        "\n"
        "    -- récupération des attributs utiles\n"
    )

    # récupération des attributs utiles
    get_variables_strings = []
    for variable in costs_config["variables"]:
        if variable["mapping"] == "value":
            var_str = "    local {0} = tonumber(way:get_value_by_key(\"{1}\")) or way:get_value_by_key(\"{1}\")\n".format(variable["name"], variable["column_name"])
            get_variables_strings.append(var_str)
        else:
            temp_var_str  = "    local {0}_tmp = tonumber(way:get_value_by_key(\"{1}\")) or way:get_value_by_key(\"{1}\")\n".format(variable["name"], variable["column_name"])
            local_var_str = "    local {}\n".format(variable["name"])
            get_variables_strings.append(temp_var_str)
            get_variables_strings.append(local_var_str)
            for key, value in variable["mapping"].items():
                cond_str = "    if {}_tmp == \"{}\" then\n".format(variable["name"], key)
                var_str  = "        {0} = tonumber(\"{1}\") or \"{1}\"\n".format(variable["name"], value)
                var_str += "    end\n"
                get_variables_strings.append(cond_str)
                get_variables_strings.append(var_str)

    # Récupération des attributs nécéssaires à la gestion des péages, ponts et tunnels.
    get_variables_strings.append("    local acces_vehicule_leger = way:get_value_by_key(\"acces_vehicule_leger\")\n")
    get_variables_strings.append("    local position_par_rapport_au_sol = tonumber(way:get_value_by_key(\"position_par_rapport_au_sol\"))\n")

    for string in get_variables_strings:
        process_way_string += string

    # pour le nom
    process_way_string += "    local way_names = way:get_value_by_key(\"way_names\")\n"

    process_way_string += (
        "    -- noms du troncon\n"
        "    if way_names and way_names ~= '' then\n"
        "        result.name = way_names\n"
        "    end\n"
    )

    # vitesse
    process_way_string += "\n    -- vitesse\n"
    process_way_string += "    result.forward_speed  = {}\n".format(output_cost["speed_value"])
    process_way_string += "    result.backward_speed = {}\n".format(output_cost["speed_value"])
    process_way_string += "\n"

    # durée
    process_way_string += "\n    -- durée\n"
    process_way_string += "    result.duration  = {}\n".format(compute_operations_string(output_cost["operations"]))
    process_way_string += "\n"

    # gestion du sens direct
    process_way_string += "    -- gestion du sens direct\n"

    direct_conditions_str = "true"
    if output_cost["direct_conditions"]:
        direct_conditions_array = output_cost["direct_conditions"].split(";")
        direct_conditions_str = " and ".join(direct_conditions_array)

    process_way_string += "    if {} then\n".format(direct_conditions_str)
    process_way_string += "        result.forward_mode  = mode.driving\n"
    # si le coût est de type distance, il faut rajouter l'attribut rate
    if output_cost["cost_type"] == 'distance':
        process_way_string += "        result.forward_rate  = 1\n"
    process_way_string += "    else\n"
    process_way_string += "        result.forward_mode  = mode.inaccessible\n"
    process_way_string += "    end\n"
    process_way_string += "\n"

    # gestion du sens inverse
    process_way_string += "    -- gestion du sens inverse\n"

    reverse_conditions_str = "true"
    if output_cost["reverse_conditions"]:
        reverse_conditions_array = output_cost["reverse_conditions"].split(";")
        reverse_conditions_str = " and ".join(reverse_conditions_array)

    process_way_string += "    if {} then\n".format(reverse_conditions_str)
    process_way_string += "        result.backward_mode  = mode.driving\n"
    # si le coût est de type distance, il faut rajouter l'attribut rate
    if output_cost["cost_type"] == 'distance':
        process_way_string += "        result.backward_rate  = 1\n"
    process_way_string += "    else\n"
    process_way_string += "        result.backward_mode  = mode.inaccessible\n"
    process_way_string += "    end\n"
    process_way_string += "\n"

    # Gestion des pếages.
    process_way_string += "    -- Gestion des pếages.\n"
    process_way_string += "    if acces_vehicule_leger == 'A péage' then\n"
    process_way_string += "        result.forward_classes[\"toll\"] = true\n"
    process_way_string += "        result.backward_classes[\"toll\"] = true\n"
    process_way_string += "    end\n"
    process_way_string += "\n"

    # Gestion des ponts.
    process_way_string += "    -- Gestion des ponts.\n"
    process_way_string += "    if position_par_rapport_au_sol and position_par_rapport_au_sol > 0 then\n"
    process_way_string += "        result.forward_classes[\"bridge\"] = true\n"
    process_way_string += "        result.backward_classes[\"bridge\"] = true\n"
    process_way_string += "    end\n"
    process_way_string += "\n"

    # Gestion des tunnels.
    process_way_string += "    -- Gestion des tunnels.\n"
    process_way_string += "    if position_par_rapport_au_sol and position_par_rapport_au_sol < 0 then\n"
    process_way_string += "        result.forward_classes[\"tunnel\"] = true\n"
    process_way_string += "        result.backward_classes[\"tunnel\"] = true\n"
    process_way_string += "    end\n"
    process_way_string += "\n"

    process_way_string += "end\n"

    return process_way_string

def _build_return():
    """
    Fonction de construction de la fonction return du fichier .lua

    Returns
    -------
    str
        chaîne de caractères correspondant à la fonction return du fichier .lua
    """

    return_string = (
        "return {\n"
        "  setup = setup,\n"
        "  process_way = process_way,\n"
        "  process_node = process_node\n"
        "}\n"
    )
    return return_string



from datetime import date
from math import ceil
import os
import time

from lxml import etree
from psycopg2.extras import DictCursor

from r2gg._osm_building import writeNode, writeWay, writeWayNds, writeRes, writeWayTags
from r2gg._sql_building import getQueryByTableAndBoundingBox
from r2gg._osm_to_pbf import osm_to_pbf


def pivot_to_osm(config, source, db_configs, connection, logger, output_is_pbf = False):
    """
    Fonction de conversion depuis la bdd pivot vers le fichier osm puis pbf le cas échéant

    Parameters
    ----------
    config: dict
        dictionnaire correspondant à la configuration décrite dans le fichier passé en argument
    source: dict
    db_configs: dict
        dictionnaire correspondant aux configurations des bdd
    connection: psycopg2.connection
        connection à la bdd de travail
    logger: logging.Logger
    """

    logger.info("Convert pivot to OSM format for a source")

    # Récupération de la date d'extraction
    work_dir_config = config['workingSpace']['directory']

    # Get extraction date from file or use current date
    date_file = os.path.join(work_dir_config, "r2gg.date")
    if os.path.exists(date_file):
        f = open(date_file, "r")
        extraction_date = f.read()
        f.close()
    else:
        extraction_date = date.today().strftime("%Y-%m-%d")

    source_db_config = db_configs[source['mapping']['source']['baseId']]
    input_schema = source_db_config.get('schema')

    cursor = connection.cursor(cursor_factory=DictCursor)

    logger.info(f"SQL: select last_value from {input_schema}.nodes_id_seq")
    cursor.execute(f"select last_value from {input_schema}.nodes_id_seq")
    vertexSequence = cursor.fetchone()[0]
    logger.info(vertexSequence)

    logger.info(f"SQL: select last_value from {input_schema}.edges_id_seq")
    cursor.execute(f"select last_value from {input_schema}.edges_id_seq")
    edgeSequence = cursor.fetchone()[0]
    logger.info(edgeSequence)

    logger.info("Starting conversion from pivot to OSM")
    start_time = time.time()

    filename = os.path.join(work_dir_config, source['id'] + ".osm")
    logger.info("OSM file to write : " + filename)

    os.makedirs(os.path.dirname(filename) or '.', exist_ok=True)
    try:
        with etree.xmlfile(filename, encoding='utf-8', close=True) as xf:
            xf.write_declaration()
            attribs = {"version": "0.6", "generator": "r2gg"}
            with xf.element("osm", attribs):

                # Ecriture des nodes
                sql_query = getQueryByTableAndBoundingBox(f'{input_schema}.nodes', source['bbox'])
                logger.info("SQL: {}".format(sql_query))
                st_execute = time.time()
                cursor.execute(sql_query)
                et_execute = time.time()
                logger.info("Execution ended. Elapsed time : %s seconds." %(et_execute - st_execute))
                row = cursor.fetchone()
                logger.info("Writing nodes")
                st_execute = time.time()
                i = 1
                while row:
                    nodeEl = writeNode(row, extraction_date)
                    xf.write(nodeEl, pretty_print=True)
                    row = cursor.fetchone()
                    if (i % ceil(cursor.rowcount/10) == 0):
                        logger.info("%s / %s nodes ajoutés" %(i, cursor.rowcount))
                    i += 1
                et_execute = time.time()
                logger.info("Writing nodes ended. Elapsed time : %s seconds." %(et_execute - st_execute))

                # Ecriture des ways
                sql_query2 = getQueryByTableAndBoundingBox(f'{input_schema}.edges', source['bbox'], ['*', f'{input_schema}.inter_nodes(geom) as internodes'])
                logger.info("SQL: {}".format(sql_query2))
                st_execute = time.time()
                cursor.execute(sql_query2)
                et_execute = time.time()
                logger.info("Execution ended. Elapsed time : %s seconds." %(et_execute - st_execute))
                row = cursor.fetchone()
                logger.info("Writing ways")
                st_execute = time.time()
                i = 1
                while row:
                    wayEl = writeWay(row, extraction_date)
                    for node in row['internodes']:
                        vertexSequence = vertexSequence + 1
                        node['id'] = vertexSequence
                        nodeEl = writeNode(node, extraction_date)
                        xf.write(nodeEl, pretty_print=True)
                    wayEl = writeWayNds(wayEl, row, row['internodes'])
                    wayEl = writeWayTags(wayEl, row)
                    xf.write(wayEl, pretty_print=True)
                    row = cursor.fetchone()
                    if (i % ceil(cursor.rowcount/10) == 0):
                        logger.info("%s / %s ways ajoutés" %(i, cursor.rowcount))
                    i += 1
                et_execute = time.time()
                logger.info("Writing ways ended. Elapsed time : %s seconds." %(et_execute - st_execute))

                # Ecriture des restrictions
                sql_query3 = f"select * from {input_schema}.non_comm"
                logger.info("SQL: {}".format(sql_query3))
                st_execute = time.time()
                cursor.execute(sql_query3)
                et_execute = time.time()
                logger.info("Execution ended. Elapsed time : %s seconds." %(et_execute - st_execute))
                row = cursor.fetchone()
                logger.info("Writing restrictions")
                st_execute = time.time()
                i = 1
                while row:
                    if row['common_vertex_id'] == -1:
                        row = cursor.fetchone()
                        i += 1
                        continue
                    ResEl = writeRes(row, i, extraction_date)
                    xf.write(ResEl, pretty_print=True)
                    row = cursor.fetchone()
                    if (i % ceil(cursor.rowcount/10) == 0):
                        logger.info("%s / %s restrictions ajoutés" %(i, cursor.rowcount))
                    i += 1
                et_execute = time.time()
                logger.info("Writing restrictions ended. Elapsed time : %s seconds." %(et_execute - st_execute))

    except etree.SerialisationError:
        logger.warn("WARNING: XML file not closed properly (lxml.etree.SerialisationError)")

    cursor.close()
    end_time = time.time()
    logger.info("Conversion from pivot to OSM ended. Elapsed time : %s seconds." %(end_time - start_time))

    # osm2pbf : Gestion du format osm.pbf
    if output_is_pbf:
        osm_to_pbf(filename, filename+'.pbf', logger)



from collections import defaultdict
from shutil import copyfile, SameFileError

def copy_file_locally(in_path, out_path):
    """
    Copie le fichies avec les chemins définis dans 'in_path' et 'out_path'

    Parameters
    ----------
    in_paths: list
        chemins sur la machine locale
    out_paths: list
        chemins sur la machine distante
    """
    try:
        copyfile(in_path, out_path)
    except SameFileError:
        print("The file " + in_path + " is already there")
    except FileNotFoundError:
        print("The file " + in_path + " was not found")



from lxml import etree

def writeNode(node, extraction_date):
    """
    Fonction qui écrit un noeud en xml à partir d'un dict (potentiellement une ligne de bdd)

    Parameters
    ----------
    node: dict
        Dictionnaire des attributs du noeud.
        pour l'instant ne contient que la clé "id"
    extraction_date: str
        date de l'extraction des données au format YYYY-MM-DD

    Returns
    -------
    lxml.etree.Element
        element xml correspondant au noeud
    """

    nodeEl = etree.Element("node", id="%s" %node['id'])

    nodeEl.set('lon', '%s' % node['lon'])
    nodeEl.set('lat', '%s' % node['lat'])
    nodeEl.set('version', '1')
    nodeEl.set('timestamp', '%sT13:37:00Z' % extraction_date)

    return nodeEl

def writeWay(way, extraction_date):
    """
    Fonction qui écrit une arrête vierge en xml à partir d'un dict
    (potentiellement une ligne de bdd)

    Parameters
    ----------
    way: dict
        Dictionnaire des attributs de l'arrête.
    extraction_date: str
        date de l'extraction des données au format YYYY-MM-DD

    Returns
    -------
    lxml.etree.Element
        element xml correspondant à l'arrête vierge
    """

    wayEl = etree.Element("way", id="%s" %way['id'])
    wayEl.set('version', '1')
    wayEl.set('timestamp', '%sT13:37:00Z' % extraction_date)

    return wayEl

def writeWayNds(wayEl, way, internodes):
    """
    Fonction qui écrit dans un arrête en xml les noeuds intermédiaires

    Parameters
    ----------
    wayEl: lxml.etree.Element
        objet xml corrrespondant à l'arrête
    way: dict
        Dictionnaire des attributs de l'arrête.
    internodes: [dict]
        Liste de dict, réponse de la fonction inter_nodes(geom) définie en SQL

    Returns
    -------
    lxml.etree.Element
        element xml correspondant à l'arrête avec ses noeuds
    """

    nd = etree.SubElement(wayEl, 'nd')
    nd.set('ref', '%s' %way['source_id'])

    for internode in internodes:
        nd = etree.SubElement(wayEl, 'nd')
        nd.set('ref', '%s' %internode['id'])

    nd = etree.SubElement(wayEl, 'nd')
    nd.set('ref', '%s' %way['target_id'])

    return wayEl

# Ecriture des restrictions
def writeRes(res, i, extraction_date):
    """
    Fonction qui écrit une restriction (non communication) en xml à partir d'un dict

    Parameters
    ----------
    res: dict
        Dictionnaire des attributs de la restriction.
    i: int
        identifiant de la retriction
    extraction_date: str
        date de l'extraction des données au format YYYY-MM-DD

    Returns
    -------
    lxml.etree.Element
        element xml correspondant à la restriction
    """

    resEl = etree.Element("relation", id="%s" %i)
    resEl.set('version', '1')
    resEl.set('timestamp', '%sT13:37:00Z' % extraction_date)

    _from = etree.SubElement(resEl, 'member')
    _from.set('type', 'way')
    _from.set('ref', '%s' %res['id_from'])
    _from.set('role','from')

    _from = etree.SubElement(resEl, 'member')
    _from.set('type', 'node')
    _from.set('ref', '%s' %res['common_vertex_id'])
    _from.set('role','via')

    _to = etree.SubElement(resEl, 'member')
    _to.set('type', 'way')
    _to.set('ref', '%s' %res['id_to'])
    _to.set('role','to')

    _tag0 = etree.SubElement(resEl, 'tag')
    _tag0.set('k','restriction')
    # TODO : Peut on définir plus précisément les restrictions ?
    _tag0.set('v','no_entry')

    _tag = etree.SubElement(resEl, 'tag')
    _tag.set('k','type')
    _tag.set('v','restriction')

    return resEl

def writeWayTags(wayEl, way):
    """
    Fonction qui ajoute des attributs (tags) à un élement xml d'arrête

    Parameters
    ----------
    wayEl: lxml.etree.Element
        objet xml corrrespondant à l'arrête
    way: int
        Dictionnaire des attributs de l'arrête.

    Returns
    -------
    lxml.etree.Element
        element xml correspondant à l'arrête avec ses attributs
    """

    for k,v in way.items():
        if k == 'internodes' or k.endswith('_id') or k == 'geom' or k=='geometrie' :
            continue
        tag = etree.SubElement(wayEl, 'tag')
        tag.set('k', '%s' %k)
        tag.set('v', '%s' %str(v))
    return wayEl



import json

def config_from_path(path_to_config_file):
    """
    Retourne une configuration (normalisée) à partir d'un chemin vers une configuration JSON

    Parameters
    ----------
    path_to_config_file: str

    Returns
    -------
    dict
        configuration normalisée
    """
    with open(path_to_config_file, encoding='utf-8') as config_file:
        raw_config = json.loads(config_file.read())

    return _normalize(raw_config)

def _normalize(raw_config):
    """
    TODO
    Normalise une configuration

    Parameters
    ----------
    raw_config: dict

    Returns
    -------
    dict
        configuration normalisée
    """
    normalized_config = raw_config
    return normalized_config



#! python3  # noqa: E265

"""Setup script to package into a Python module."""

# ############################################################################
# ########## Libraries #############
# ##################################

# standard library
from pathlib import Path

# 3rd party
from setuptools import find_packages, setup

# package (to get version)
from r2gg import __about__

# ############################################################################
# ########### Globals ##############
# ##################################

# The directory containing this file
HERE = Path(__file__).parent

requirements = []
with open(HERE / "requirements/base.txt") as f:
    requirements = [
        line
        for line in f.read().splitlines()
        if not line.startswith(("#", "-")) and len(line)
    ]

# The text of the README file
README = (HERE / "README.md").read_text()

# ############################################################################
# ########### Main #################
# ##################################

setup(
    name=__about__.__package_name__,
    author=__about__.__author__,
    author_email=__about__.__email__,
    description=__about__.__summary__,
    long_description=README,
    long_description_content_type="text/markdown",
    keywords=__about__.__keywords__,
    url=__about__.__uri__,
    version=__about__.__version__,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Information Technology",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: GIS",
    ],
    # packaging
    packages=find_packages(
        exclude=["contrib", "docs", "*.tests", "*.tests.*", "tests.*", "tests", ".venv"]
    ),
    include_package_data=True,
    install_requires=requirements,

    # cli
    entry_points={
        "console_scripts": [
            f"{__about__.__executable_name__} = r2gg.cli:main",
            f"{__about__.__executable_name__}-sql2pivot = r2gg.cli:sql2pivot",
            f"{__about__.__executable_name__}-pivot2pgrouting = r2gg.cli:pivot2pgrouting",
            f"{__about__.__executable_name__}-pivot2osm = r2gg.cli:pivot2osm",
            f"{__about__.__executable_name__}-osm2osrm = r2gg.cli:osm2osrm",
            f"{__about__.__executable_name__}-osm2valhalla = r2gg.cli:osm2valhalla",
            f"{__about__.__executable_name__}-road2config = r2gg.cli:road2config"
        ]
    },
)



import argparse
import logging
import os

import psycopg2

from r2gg._read_config import config_from_path

# Définition des niveaux de log
LEVELS = {
    'CRITICAL' : logging.CRITICAL,
    'ERROR' : logging.ERROR,
    'WARNING' : logging.WARNING,
    'INFO' : logging.INFO,
    'DEBUG' : logging.DEBUG
}

def configure():
    """
    Fonction de lecture du fichier de configuration passé en argument

    Returns
    -------
    config: dict
        dictionnaire correspondant à la configuration décrite dans le fichier passé en argument
    resource: dict
        dictionnaire correspondant à la resource décrite dans le fichier passé en argument
    db_configs: dict
        dictionnaire correspondant aux configurations des bdd
    connection: psycopg2.connection
        connection à la bdd de travail
    logger: logging.Logger
    """
    parser = argparse.ArgumentParser()
    parser.add_argument('config_file_path', type=str)
    args = parser.parse_args()
    config_path = args.config_file_path

    # Récupération de l'objet 'génération' qui contient toute la config
    config = config_from_path(config_path)['generation']

    # Récupération de la configuration du log
    logs_config = config_from_path( config['general']['logs']['configFile'] )

    # Gestion du fichiers de logs non spécifié
    try:
        logs_file = logs_config['filename']
    except KeyError:
        logs_file = '/dev/null'

    # Configuration du module logging
    logging.basicConfig(
        format='%(asctime)s %(message)s',
        level=LEVELS[ logs_config['level'].upper() ],
        handlers=[
            logging.FileHandler(logs_file),
            logging.StreamHandler()
        ])

    # Initialisation du logger
    logger = logging.getLogger(__name__)
    logger.info("Log initialized")

    # Todo : Créer une fonction qui vérifie la configuration
    db_configs = {}
    # Configuration des bases de données précisées dans la config
    for base in config['bases']:
        if base['type'] == 'bdd':
            db_configs[ base['id'] ] = config_from_path(base['configFile'])
            db_configs[base['id']].update({"schema":base['schema']})

    # Récupération de l'objet permettant de générer la ressource
    resource = config['resource']

    # Création de l'espace de travail
    if not os.path.exists(config['workingSpace']['directory']):
        os.makedirs(config['workingSpace']['directory'])

    return config, resource, db_configs, logger

def connect_working_db(config, db_configs, logger):
    """
    Fonction de connexion à la BDD de travail

    Parameters
    ----------
    config: dict
        dictionnaire correspondant à la configuration décrite dans le fichier passé en argument
    db_configs: dict
        dictionnaire correspondant aux configurations des bdd
    logger: logging.Logger
    Returns
    -------
    connection: psycopg2.connection
        connection à la bdd de travail

    """

    # Configuration de la bdd de travail
    work_db_config = db_configs[ config['workingSpace']['baseId'] ]

    # Récupération des paramètres de la bdd
    host = work_db_config.get('host')
    dbname = work_db_config.get('database')
    user = work_db_config.get('user')
    password = work_db_config.get('password')
    port = work_db_config.get('port')
    connect_args = 'host=%s dbname=%s user=%s password=%s port=%s' %(host, dbname, user, password, port)

    logger.info("Connecting to work database")
    connection = psycopg2.connect(connect_args)
    connection.set_client_encoding('UTF8')

    return connection

def disconnect_working_db(connection, logger):
    """
    Fonction de connexion à la BDD de travail

    Parameters
    ----------
    connection: psycopg2.connection
        connection à la bdd de travail
    logger: logging.Logger
    """

    connection.close()
    logger.info("Connection to work database closed")


#! python3  # noqa: E265

"""Metadata about the package to easily retrieve informations about it.

See: https://packaging.python.org/guides/single-sourcing-package-version/
"""

from datetime import date

__all__ = [
    "__author__",
    "__copyright__",
    "__email__",
    "__license__",
    "__summary__",
    "__title__",
    "__uri__",
    "__version__",
]

__author__ = "IGNF"
__copyright__ = "2022 - {0}, {1}".format(date.today().year, __author__)
__email__ = ""
__executable_name__ = "r2gg"
__package_name__ = "r2gg"

__keywords__ = ['cli', ' IGN', ' osrm', ' routing', ' pgrouting', ' valhalla', ' generation-algorithms', ' isochrone', ' road2']
__license__ = "GPL v3"
__summary__ = "Route Graph Generator (r2gg) est un script Python qui permet la génération de graphes pour des moteurs de calcul d'itinéraire"
__title__ = "Route Graph Generator"
__title_clean__ = "".join(e for e in __title__ if e.isalnum())
__uri_homepage__ = "https://ignf.github.io/route-graph-generator"
__uri_repository__ = "https://github.com/IGNF/route-graph-generator/"
__uri_tracker__ = f"{__uri_repository__}issues/"
__uri__ = __uri_repository__

__version__ = "2.2.2"
__version_info__ = tuple(
    [
        int(num) if num.isdigit() else num
        for num in __version__.replace("-", ".", 1).split(".")
    ]
)


__cli_usage__ = f"{__summary__}."



def build_valhalla_lua(costs_config):
    """
    Fonction qui crée un fichier .lua de graphe Valhalla à partir d'un dictionnaire de
    définition des coûts (costs_config)

    Parameters
    ----------
    costs_config: dict
        Dictionnaire de configuration des coûts correspondant au schéma JSON de
        configuration des coûts

    Returns
    -------
    str
        chaîne de caractères correspondant à un fichier .lua de graphe Valhalla
    """

    # Récupération du nom du champ de vitesse pour les voitures
    auto_speed_name = "vitesse_moyenne_vl"
    for variable in costs_config["variables"]:
        if variable["name"] == "vitesse_voiture":
            auto_speed_name = variable["column_name"]

    # TODO: ajouter plus de variables issues de costs_config (quand cost_profile sera ajouté à la configuration des coûts)

    result = """
way_type = {{
["no_pedestrian"] =  {{["auto_forward"] = "true", ["pedestrian_forward"] = "false"}},
["regular"] =        {{["auto_forward"] = "true", ["pedestrian_forward"] = "true"}},
["no_car"] =         {{["auto_forward"] = "false", ["pedestrian_forward"] = "true"}}
}}

restriction = {{
["no_left_turn"] = 0,
["no_right_turn"] = 1,
["no_straight_on"] = 2,
["no_u_turn"] = 3,
["only_right_turn"] = 4,
["only_left_turn"] = 5,
["only_straight_on"] = 6,
["no_entry"] = 7,
["no_exit"] = 8,
["no_turn"] = 9
}}

function round(val, n)
  if (n) then
    return math.floor( (val * 10^n) + 0.5) / (10^n)
  else
    return math.floor(val+0.5)
  end
end

function restriction_prefix(restriction_str)
  --not a string
  if restriction_str == nil then
    return nil
  end

  --find where the restriction type ends.
  --format looks like
  --restriction:conditional=no_left_turn @ (07:00-09:00,15:30-17:30)
  local index = 0
  local found = false
  for c in restriction_str:gmatch"." do
    if c == "@" then
      found = true
      break
    end
    if c ~= " " then
      index = index + 1
    end
  end

  --@ symbol did not exist
  if found == false then
    return nil
  end

  --return the type of restriction
  return restriction_str:sub(0, index)
end

function restriction_suffix(restriction_str)
  --not a string
  if restriction_str == nil then
    return nil
  end

  --find where the restriction type ends.
  --format looks like
  --restriction:conditional=no_left_turn @ (07:00-09:00,15:30-17:30)
  local index = 0
  local found = false
  for c in restriction_str:gmatch"." do

    if found == true then
      if c ~= " " then
        index = index + 1
        break
      end
    elseif c == "@" then
      found = true
    end
    index = index + 1
  end

  --@ symbol did not exist
  if found == false then
    return nil
  end

  --return the date and time information for the restriction
  return restriction_str:sub(index, string.len(restriction_str))
end


--convert the numeric (non negative) number portion at the beginning of the string
function numeric_prefix(num_str, allow_decimals)
  --not a string
  if num_str == nil then
    return nil
  end

  --find where the numbers stop
  local index = 0
  -- flag to say if we've seen a decimal dot already. we shouldn't allow two,
  -- otherwise the call to tonumber() might fail.
  local seen_dot = false
  for c in num_str:gmatch"." do
    if tonumber(c) == nil then
      if c == "." then
        if allow_decimals == false or seen_dot then
           break
        end
        seen_dot = true
      else
        break
      end
    end
    index = index + 1
  end

  --there weren't any numbers
  if index == 0 then
    return nil
  end

  --Returns the substring that's numeric at the start of num_str
  return num_str:sub(0, index)
end

--normalize a speed value
function normalize_speed(speed)
  --grab the number prefix
  local num = tonumber(numeric_prefix(speed, false))

  --check if the rest of the string ends in "mph" convert to kph
  if num then
    if speed:sub(-3) == "mph" then
      num = round(num * 1.609344)
    end

    --if num > 150kph or num < 10kph....toss
    if num > 150 or num < 10 then
      return nil
    end
  end

  return num
end


--returns 1 if you should filter this way 0 otherwise
function filter_tags_generic(kv)

  local forward = way_type["regular"]
  if (tonumber(kv["{0}"]) <= 0) then
    forward = way_type["no_car"]
  elseif (kv["nature"] == "Type autoroutier" or kv["nature"] == "Bretelle") then
    forward = way_type["no_pedestrian"]
  end

  local access = "true"

  kv["emergency_forward"] = "false"
  kv["emergency_backward"] = "false"

  if forward then
    for k,v in pairs(forward) do
      kv[k] = v
    end
  end

  kv["pedestrian_backward"] = kv["pedestrian_forward"]

  if (kv["direction"] ~= "0") then
    oneway_norm = "true"
  else
    oneway_norm = "false"
  end

  kv["oneway"] = oneway_norm

  if (tonumber(kv["direction"]) <= 0) then
    kv["auto_backward"] = "true"
  else
    kv["auto_backward"] = "false"
  end
  if (tonumber(kv["direction"]) >= 0) then
    kv["auto_forward"] = "true"
  else
    kv["auto_forward"] = "false"
  end

  kv["use"] = 0

  kv["road_class"] = tonumber(kv["importance"]) - 1

  local r_shoulder = nil
  local l_shoulder = r_shoulder

  kv["shoulder_right"] = r_shoulder
  kv["shoulder_left"] = l_shoulder


  local cycle_lane_right_opposite = "false"
  local cycle_lane_left_opposite = "false"

  local cycle_lane_right = 0
  local cycle_lane_left = 0

  kv["cycle_lane_right"] = cycle_lane_right
  kv["cycle_lane_left"] = cycle_lane_left

  kv["cycle_lane_right_opposite"] = cycle_lane_right
  kv["cycle_lane_left_opposite"] = cycle_lane_left

  local highway_type = kv["highway"]
  if kv["highway"] == "construction" then
    highway_type = kv["construction"]
  end

  kv["link"] = "false"

  kv["private"] = "false"
  kv["no_thru_traffic"] = "false"
  kv["ferry"] = "false"
  kv["rail"] = "false"
  kv["name"] = kv["way_names"]
  kv["name:en"] = kv["name:en"]
  kv["alt_name"] = kv["alt_name"]
  kv["official_name"] = kv["official_name"]

  local corrected_speed = kv["{0}"]
  if (kv["urbain"] == "True") then
    corrected_speed = tostring(3.6 * kv["length_m"] / ((3.6 / tonumber(corrected_speed)) * kv["length_m"] + 5))
  end

  kv["max_speed"] = corrected_speed

  kv["advisory_speed"] = normalize_speed(corrected_speed)
  kv["average_speed"] = normalize_speed(corrected_speed)
  if (tonumber(kv["direction"]) <= 0) then
    kv["backward_speed"] = normalize_speed(corrected_speed)
  else
    kv["backward_speed"] = normalize_speed("0")
  end
  if (tonumber(kv["direction"]) >= 0) then
    kv["forward_speed"] = normalize_speed(corrected_speed)
  else
    kv["forward_speed"] = normalize_speed("0")
  end

  kv["default_speed"] = normalize_speed(corrected_speed)

  if tonumber(corrected_speed) <= 0 then
    kv["auto_backward"] = "false"
    kv["auto_forward"] = "false"
  end

  kv["int"] = kv["int"]
  kv["int_ref"] = kv["int_ref"]
  kv["surface"] = kv["surface"]
  kv["wheelchair"] = kv["wheelchair"]

  if (tonumber(kv["nombre_de_voies"])) then
    lane_count = kv["nombre_de_voies"]
  else
    lane_count = nil
  end

  kv["lanes"] = lane_count

  kv["forward_lanes"] = nil
  kv["backward_lanes"] = nil
  if (tonumber(kv["position_par_rapport_au_sol"]) > 0) then
    kv["bridge"] = "true"
  else
    kv["bridge"] = "false"
  end

  kv["hov_tag"] = "true"
  if (kv["hov"] and kv["hov"] == "no") then
    kv["hov_forward"] = "false"
    kv["hov_backward"] = "false"
  else
    kv["hov_forward"] = kv["auto_forward"]
    kv["hov_backward"] = kv["auto_backward"]
  end

  if (tonumber(kv["position_par_rapport_au_sol"]) < 0) then
    kv["tunnel"] = "true"
  else
    kv["tunnel"] = "false"
  end

  if (kv["acces_vehicule_leger"] == 'A péage') then
    kv["toll"] = "true"
  else
    kv["toll"] = "false"
  end

  kv["destination"] = kv["destination"]
  kv["destination:forward"] = kv["destination:forward"]
  kv["destination:backward"] = kv["destination:backward"]
  kv["destination:ref"] = kv["destination:ref"]
  kv["destination:ref:to"] = kv["destination:ref:to"]
  kv["destination:street"] = kv["destination:street"]
  kv["destination:street:to"] = kv["destination:street:to"]
  kv["junction:ref"] = kv["junction:ref"]
  kv["turn:lanes"] = kv["turn:lanes"]
  kv["turn:lanes:forward"] = kv["turn:lanes:forward"]
  kv["turn:lanes:backward"] = kv["turn:lanes:backward"]

  if (kv["restriction_de_hauteur"] ~= "None") then
    kv["maxheight"] = kv["restriction_de_hauteur"]
  end

  if (kv["restriction_de_largeur"] ~= "None") then
    kv["maxwidth"] = kv["restriction_de_largeur"]
  end

  if (kv["restriction_de_longueur"] ~= "None") then
    kv["maxlength"] = kv["restriction_de_longueur"]
  end

  if (kv["restriction_de_poids_total"] ~= "None") then
    kv["maxweight"] = kv["restriction_de_poids_total"]
  end

  if (kv["restriction_de_poids_par_essieu"] ~= "None") then
    kv["maxaxleload"] = kv["restriction_de_poids_par_essieu"]
  end

  kv["hazmat"] = nil
  kv["maxspeed:hgv"] = normalize_speed(corrected_speed)

  if (kv["hgv:national_network"] or kv["hgv:state_network"] or kv["hgv"] == "local" or kv["hgv"] == "designated") then
    kv["truck_route"] = "true"
  end

  local nref = kv["ncn_ref"]
  local rref = kv["rcn_ref"]
  local lref = kv["lcn_ref"]
  local bike_mask = 0
  if nref or kv["ncn"] == "yes" then
    bike_mask = 1
  end
  if rref or kv["rcn"] == "yes" then
    bike_mask = bit.bor(bike_mask, 2)
  end
  if lref or kv["lcn"] == "yes" then
    bike_mask = bit.bor(bike_mask, 4)
  end
  if kv["mtb"] == "yes" then
    bike_mask = bit.bor(bike_mask, 8)
  end

  kv["bike_national_ref"] = nref
  kv["bike_regional_ref"] = rref
  kv["bike_local_ref"] = lref
  kv["bike_network_mask"] = bike_mask

  return 0
end

function nodes_proc (kv, nokeys)

  local auto = 1
  local truck = 8
  local bus = 64
  local taxi = 32
  local foot = 2
  local wheelchair = 256
  local bike = 4
  local emergency = 16
  local hov = 128
  local moped = 512
  local motorcycle = 1024

  local gate = false
  local bollard = false
  local sump_buster = false

  kv["gate"] = tostring(gate)
  kv["bollard"] = tostring(bollard)
  kv["sump_buster"] = tostring(sump_buster)

  kv["private"] = "false"

  kv["access_mask"] = bit.bor(auto, emergency, truck, bike, foot, wheelchair, bus, hov, moped, motorcycle, taxi)

  kv["tagged_access"] = 0

  return 0, kv
end

function ways_proc (kv, nokeys)
  if nokeys == 0 then
    return 1, kv, 0, 0
  end

  filter = filter_tags_generic(kv)

  return filter, kv, 0, 0
end

function rels_proc (kv, nokeys)
  if (kv["type"] == "connectivity") then
    return 0, kv
  end

  if (kv["type"] == "route" or kv["type"] == "restriction") then
     if kv["restriction:probable"] then
       if kv["restriction"] or kv["restriction:conditional"] then
         kv["restriction:probable"] = nil
       end
     end

     local restrict = restriction[kv["restriction"]] or restriction[restriction_prefix(kv["restriction:conditional"])] or
                      restriction[restriction_prefix(kv["restriction:probable"])]

     local restrict_type = restriction[kv["restriction:hgv"]] or restriction[kv["restriction:emergency"]] or
                           restriction[kv["restriction:taxi"]] or restriction[kv["restriction:motorcar"]] or
                           restriction[kv["restriction:bus"]] or restriction[kv["restriction:bicycle"]] or
                           restriction[kv["restriction:hazmat"]] or restriction[kv["restriction:motorcycle"]] or
			   restriction[kv["restriction:foot"]]

     if restrict_type ~= nil then
       restrict = restrict_type
     end

     if kv["type"] == "restriction" or kv["restriction:conditional"] or kv["restriction:probable"] then

       if restrict ~= nil then

         kv["restriction:conditional"] = restriction_suffix(kv["restriction:conditional"])
         kv["restriction:probable"] = restriction_suffix(kv["restriction:probable"])

         kv["restriction:hgv"] = restriction[kv["restriction:hgv"]]
         kv["restriction:emergency"] = restriction[kv["restriction:emergency"]]
         kv["restriction:taxi"] = restriction[kv["restriction:taxi"]]
         kv["restriction:motorcar"] = restriction[kv["restriction:motorcar"]]
         kv["restriction:motorcycle"] = restriction[kv["restriction:motorcycle"]]
         kv["restriction:bus"] = restriction[kv["restriction:bus"]]
         kv["restriction:bicycle"] = restriction[kv["restriction:bicycle"]]
         kv["restriction:hazmat"] = restriction[kv["restriction:hazmat"]]
         kv["restriction:foot"] = restriction[kv["restriction:foot"]]

         if restrict_type == nil then
           kv["restriction"] = restrict
         else
           kv["restriction"] = nil
         end

       else
         return 1, kv
       end
       return 0, kv
     elseif kv["route"] == "bicycle" or kv["route"] == "mtb" then

       local bike_mask = 0

       if kv["network"] == "mtb" or kv["route"] == "mtb" then
         bike_mask = 8
       end

       if kv["network"] == "ncn" then
         bike_mask = bit.bor(bike_mask, 1)
       elseif kv["network"] == "rcn" then
         bike_mask = bit.bor(bike_mask, 2)
       elseif kv["network"] == "lcn" then
         bike_mask = bit.bor(bike_mask, 4)
       end

       kv["bike_network_mask"] = bike_mask

       kv["day_on"] = nil
       kv["day_off"] = nil
       kv["restriction"] = nil

       return 0, kv
     elseif restrict ~= nil then
       return 1, kv
     else
       kv["day_on"] = nil
       kv["day_off"] = nil
       kv["restriction"] = nil
       return 0, kv
     end
  end

  return 1, kv
end

function rel_members_proc (keyvalues, keyvaluemembers, roles, membercount)
  membersuperseeded = {{0}}
  for i = 1, membercount do
    membersuperseeded[i] = 0
  end

  return 1, keyvalues, membersuperseeded, 0, 0, 0
end
    """.format(auto_speed_name)

    return result
